/*
 * Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */
#include "bdev_nvme.h"

#include "spdk/json.h"
#include "spdk/likely.h"
#include "spdk/bdev_module.h"
#include "spdk/nvme.h"

#include "spdk_internal/bdev_stat.h"
#include "bdev_nvme_self.h"
#include "common.h"
#include <securec.h>

void bdev_update_ch_timeout(struct nvme_bdev_poll_group *group)
{
	uint64_t current_ticks = 0;
	uint64_t poll_ticks = 0;
	int64_t poll_time = 0;

	current_ticks = spdk_get_ticks();

	if (spdk_unlikely(g_polltime_threshold)) {
		if (group->save_start_ticks) {
			poll_ticks = current_ticks - group->save_start_ticks;
			poll_time = (poll_ticks * 1000ULL) / spdk_get_ticks_hz();
			if (poll_time >= g_polltime_threshold) {
				group->num_poll_timeout++;
				SPDK_NOTICELOG("group[%p] poll timeout in %ldms", group, poll_time);
			}
		}
		group->save_start_ticks = current_ticks;
	}
}

int _bdev_nvme_submit_request_self(struct spdk_io_channel *ch, struct spdk_bdev_io *bdev_io)
{
	struct nvme_io_channel *nvme_ch = spdk_io_channel_get_ctx(ch);

	if (nvme_ch->qpair == NULL) {
		/* The device is currently resetting */
		return -1;
	}

	switch (bdev_io->type) {
	case SPDK_BDEV_IO_TYPE_READ_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "read %lu lbas with offset %#lx\n", bdev_io->u.contig.num_blocks,
			      bdev_io->u.contig.offset_blocks);
		return bdev_nvme_queue_cmd_with_md((struct nvme_bdev *)bdev_io->bdev->ctxt, nvme_ch->qpair,
						   bdev_io->driver_ctx, bdev_io->u.contig.buf,
						   bdev_io->u.contig.md_buf, SPDK_BDEV_IO_TYPE_READ,
						   bdev_io->u.contig.num_blocks, bdev_io->u.contig.offset_blocks);
	case SPDK_BDEV_IO_TYPE_WRITE_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "write %lu lbas with offset %#lx\n", bdev_io->u.contig.num_blocks,
			      bdev_io->u.contig.offset_blocks);
		return bdev_nvme_queue_cmd_with_md((struct nvme_bdev *)bdev_io->bdev->ctxt, nvme_ch->qpair,
						   bdev_io->driver_ctx, bdev_io->u.contig.buf,
						   bdev_io->u.contig.md_buf, SPDK_BDEV_IO_TYPE_WRITE,
						   bdev_io->u.contig.num_blocks, bdev_io->u.contig.offset_blocks);
	case SPDK_BDEV_IO_TYPE_READV_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "readv %lu lbas with offset %#lx\n", bdev_io->u.bdev.num_blocks,
			      bdev_io->u.bdev.offset_blocks);
		return bdev_nvme_queue_cmd_v_with_md((struct nvme_bdev *)bdev_io->bdev->ctxt, nvme_ch->qpair,
						     bdev_io->driver_ctx, SPDK_BDEV_IO_TYPE_READ,
						     bdev_io->u.bdev.iovs, bdev_io->u.bdev.iovcnt,
						     bdev_io->u.bdev.num_blocks, bdev_io->u.bdev.offset_blocks);
	case SPDK_BDEV_IO_TYPE_WRITEV_NVME:
		SPDK_DEBUGLOG(bdev_nvme, "writev %lu lbas with offset %#lx\n", bdev_io->u.bdev.num_blocks,
			      bdev_io->u.bdev.offset_blocks);
		return bdev_nvme_queue_cmd_v_with_md((struct nvme_bdev *)bdev_io->bdev->ctxt, nvme_ch->qpair,
						     bdev_io->driver_ctx, SPDK_BDEV_IO_TYPE_WRITE,
						     bdev_io->u.bdev.iovs, bdev_io->u.bdev.iovcnt,
						     bdev_io->u.bdev.num_blocks, bdev_io->u.bdev.offset_blocks);
	case SPDK_BDEV_IO_TYPE_UNMAP_BLOCKS:
		return bdev_nvme_unmap_blocks((struct nvme_bdev *)bdev_io->bdev->ctxt,
					      ch,
					      (void *)bdev_io->driver_ctx,
					      (struct spdk_nvme_dsm_range *)bdev_io->u.contig.buf,
					      bdev_io->u.contig.num_blocks);
	default:
		return -EINVAL;
	}
	return 0;
}

int bdev_nvme_dump_info_json_self(void *ctx, struct spdk_json_write_ctx *w)
{
	return 0;
}

uint16_t bdev_nvme_get_io_channel_id(struct spdk_io_channel *ch)
{
	struct nvme_io_channel *nvme_ch = spdk_io_channel_get_ctx(ch);
	uint16_t channel_id;
	struct spdk_nvme_qpair	*qpair = nvme_ch->qpair;
	channel_id = spdk_nvme_get_qpair_id(qpair);
	return channel_id;
}

uint64_t bdev_nvme_get_timeout_count(struct spdk_io_channel *ch)
{
	struct nvme_io_channel *nvme_ch = spdk_io_channel_get_ctx(ch);
	return nvme_ch->group->num_poll_timeout;
}

int32_t nvme_ctrlr_get_info(const char *ctrlName, struct nvme_ctrlr_info **ppCtrlr)
{
	uint32_t num_ctrlr = 0, i = 0;
	struct nvme_bdev_ctrlr *nvme_bdev_ctrlr = NULL;
	struct nvme_ctrlr_info *pCtrlrInfo = NULL;
	const struct spdk_nvme_ctrlr_data *cdata = NULL;
	struct spdk_nvme_ctrlr_opts *opts = NULL;

	struct spdk_pci_device *pci_dev = NULL;
	int rc;

	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		num_ctrlr++;
	}
	if (num_ctrlr == 0) {
		SPDK_NOTICELOG("No any nvme controller.\n");
		return 0;
	}
	if (ctrlName != NULL) {
		num_ctrlr = 1;
	}
	pCtrlrInfo = calloc(num_ctrlr, sizeof(struct nvme_ctrlr_info));
	if (pCtrlrInfo == NULL) {
		SPDK_ERRLOG("Failed to alloc memory for getting controller infomation.\n");
		return -1;
	}
	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		if (i >= num_ctrlr) { /* prevent having controllers be added or deleted */
			i++;
			continue;
		}
		if (ctrlName != NULL) {
			if (strcmp(nvme_bdev_ctrlr->name, ctrlName) != 0) {
				continue;
			}
		}
		cdata = spdk_nvme_ctrlr_get_data(nvme_bdev_ctrlr->ctrlr);
		opts = spdk_nvme_ctrlr_get_opts(nvme_bdev_ctrlr->ctrlr);
		pci_dev = spdk_nvme_ctrlr_get_pci_device(nvme_bdev_ctrlr->ctrlr);
		if (pci_dev == NULL) {
			SPDK_ERRLOG("Failed to get pci device\n");
			break;
		}
		rc = strcpy_s(pCtrlrInfo[i].ctrlName, sizeof(pCtrlrInfo[i].ctrlName), nvme_bdev_ctrlr->name);
		if (rc != 0) {
			SPDK_ERRLOG("String copy failed\n");
		}
		rc = strcpy_s(pCtrlrInfo[i].pciAddr, sizeof(pCtrlrInfo[i].pciAddr),
			      nvme_bdev_ctrlr->connected_trid->traddr);
		if (rc != 0) {
			SPDK_ERRLOG("String copy failed\n");
		}

		rc = memcpy_s(pCtrlrInfo[i].sn, sizeof(pCtrlrInfo[i].sn), cdata->sn, 20);
		if (rc != 0) {
			SPDK_ERRLOG("Memory copy failed\n");
		}

		rc = memcpy_s(pCtrlrInfo[i].fr, sizeof(pCtrlrInfo[i].fr), cdata->fr, 8);
		if (rc != 0) {
			SPDK_ERRLOG("Memory copy failed\n");
		}

		rc = memcpy_s(pCtrlrInfo[i].mn, sizeof(pCtrlrInfo[i].mn), cdata->mn, 40);
		if (rc != 0) {
			SPDK_ERRLOG("Memory copy failed\n");
		}

		pCtrlrInfo[i].trtype = (uint16_t)nvme_bdev_ctrlr->connected_trid->trtype;
		pCtrlrInfo[i].tnvmcap = cdata->tnvmcap[0];
		pCtrlrInfo[i].unvmcap = cdata->unvmcap[0];
		pCtrlrInfo[i].support_ns = cdata->oacs.ns_manage;
		pCtrlrInfo[i].directives = cdata->oacs.directives;
		pCtrlrInfo[i].dsm = cdata->oncs.dsm;
		pCtrlrInfo[i].max_num_ns = cdata->nn;
		pCtrlrInfo[i].num_io_queues = opts->num_io_queues;
		pCtrlrInfo[i].io_queue_size = opts->io_queue_size;
		pCtrlrInfo[i].device_id = spdk_pci_device_get_device_id(pci_dev);
		pCtrlrInfo[i].subdevice_id = spdk_pci_device_get_subdevice_id(pci_dev);
		pCtrlrInfo[i].vid = cdata->vid;
		pCtrlrInfo[i].ssvid = cdata->ssvid;
		pCtrlrInfo[i].ctrlid = cdata->cntlid;
		pCtrlrInfo[i].version = spdk_nvme_ctrlr_get_regs_vs(nvme_bdev_ctrlr->ctrlr).raw;
		i++;
		if (ctrlName != NULL) {
			break;
		}
	}
	if (i != num_ctrlr) {
		SPDK_ERRLOG("It has controller been added or deleted when fetched infomation, please try again later.\n");
		free(pCtrlrInfo);
		return -1;
	}
	*ppCtrlr = pCtrlrInfo;
	return num_ctrlr;
}

struct nvme_bdev_ctrlr *nvme_ctrlr_get_by_name(const char *name)
{
	struct nvme_bdev_ctrlr *nvme_bdev_ctrlr = NULL;

	if (name == NULL) {
		return NULL;
	}

	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		if (strcmp(name, nvme_bdev_ctrlr->name) == 0) {
			return nvme_bdev_ctrlr;
		}
	}

	return NULL;
}

struct spdk_nvme_ctrlr *spdk_nvme_ctrlr_get_by_name(const char *ctrlname)
{
	struct nvme_bdev_ctrlr	*nvme_bdev_ctrlr = NULL;

	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		if (strcmp(nvme_bdev_ctrlr->name, ctrlname) == 0) {
			return nvme_bdev_ctrlr->ctrlr;
		}
	}

	return NULL;
}

struct spdk_nvme_ctrlr *spdk_nvme_ctrlr_get_by_ctrlr(const struct nvme_bdev_ctrlr *nvme_bdev_ctrlr)
{
	if (nvme_bdev_ctrlr == NULL) {
		return NULL;
	}
	return nvme_bdev_ctrlr->ctrlr;
}

void nvme_ctrlr_clear_iostat_by_name(const char *ctrlname)
{
	int i;
	size_t size = strnlen(ctrlname, 24);

	for (i = 0; i < STAT_MAX_NUM; i++) {
		if (strncmp(g_io_stat_map[i].bdev_name, ctrlname, size) == 0) {
			if ((g_io_stat_map[i].bdev_name[size] == 'n') && isdigit(g_io_stat_map[i].bdev_name[size + 1])) {
				g_io_stat_map[i].channel_id = 0;
				memset(g_io_stat_map[i].bdev_name, 0, sizeof(g_io_stat_map[i].bdev_name));
				g_io_stat_map[i].num_read_ops = 0;
				g_io_stat_map[i].num_write_ops = 0;
				g_io_stat_map[i].bytes_read = 0;
				g_io_stat_map[i].bytes_written = 0;
				g_io_stat_map[i].io_outstanding = 0;
				g_io_stat_map[i].read_latency_ticks = 0;
				g_io_stat_map[i].write_latency_ticks = 0;
				g_io_stat_map[i].io_ticks = 0;
				/* used flag set false in last avoid race in channel create */
				g_io_stat_map[i].used = false;
			}
		}
	}
}

void nvme_ctrlr_clear_iostat_all(void)
{
	struct nvme_bdev_ctrlr	*nvme_bdev_ctrlr = NULL;

	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		nvme_ctrlr_clear_iostat_by_name(nvme_bdev_ctrlr->name);
	}
}

struct spdk_nvme_ns *bdev_nvme_get_ns(struct nvme_bdev *nbdev)
{
	return nbdev->nvme_ns->ns;
}

void bdev_nvme_update_block_by_nvme_ctrlr(struct spdk_nvme_ctrlr *ctrlr)
{
	uint32_t i;
	struct nvme_bdev_ctrlr *nvme_bdev_ctrlr = NULL;
	struct nvme_bdev_ns	*ns = NULL;
	struct nvme_bdev	*nvme_bdev = NULL, *tmp = NULL;


	pthread_mutex_lock(&g_bdev_nvme_mutex);
	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		if (nvme_bdev_ctrlr->ctrlr != ctrlr) {
			continue;
		}

		pthread_mutex_unlock(&g_bdev_nvme_mutex);
		for (i = 0; i < nvme_bdev_ctrlr->num_ns; i++) {
			ns = nvme_bdev_ctrlr->namespaces[i];
			TAILQ_FOREACH_SAFE(nvme_bdev, &ns->bdevs, tailq, tmp) {
				nvme_bdev->disk.blocklen = spdk_nvme_ns_get_sector_size(nvme_bdev->nvme_ns->ns);
				nvme_bdev->disk.blockcnt = spdk_nvme_ns_get_num_sectors(nvme_bdev->nvme_ns->ns);
			}
		}
		return;
	}
	pthread_mutex_unlock(&g_bdev_nvme_mutex);
}

int bdev_nvme_update_ns(struct nvme_bdev_ctrlr *nvme_bdev_ctrlr, uint32_t nsid)
{
	struct nvme_bdev_ns	*ns = NULL;
	bool ns_active = spdk_nvme_ctrlr_is_active_ns(nvme_bdev_ctrlr->ctrlr, nsid);

	if (nvme_bdev_ctrlr == NULL || nsid > nvme_bdev_ctrlr->num_ns) {
		SPDK_ERRLOG("Parameter error. nsid[%u], the max nsid is[%u]\n", nsid, nvme_bdev_ctrlr->num_ns);
		return -1;
	}

	ns = nvme_bdev_ctrlr->namespaces[nsid - 1];
	ns->type = NVME_BDEV_NS_STANDARD;

	if (!ns->populated && ns_active) {
		SPDK_NOTICELOG("NSID %u to be added\n", nsid);
		ns->id = nsid;
		ns->ctrlr = nvme_bdev_ctrlr;
		TAILQ_INIT(&ns->bdevs);
		/* add a new bdev device in this ns */
		nvme_ctrlr_populate_namespace(nvme_bdev_ctrlr, ns, NULL);
		return 0;
	}

	if (ns->populated && !ns_active) {
		SPDK_NOTICELOG("NSID %u is removed\n", nsid);
		nvme_ctrlr_depopulate_namespace(nvme_bdev_ctrlr, ns);
		return 0;
	}

	return 0;
}

bool spdk_bdev_can_remove(struct nvme_bdev_ctrlr *nvme_bdev_ctrlr, uint32_t nsid)
{
	struct nvme_bdev_ns *ns = NULL;
	struct nvme_bdev *bdev = NULL, *tmp = NULL;
	bool empty = false;

	ns = nvme_bdev_ctrlr->namespaces[nsid - 1];
	if (ns == NULL) {
		return true;
	}

	TAILQ_FOREACH_SAFE(bdev, &ns->bdevs, tailq, tmp) {
		pthread_mutex_lock(&bdev->disk.internal.mutex);
		empty = TAILQ_EMPTY(&bdev->disk.internal.open_descs);
		/* for each bdev in ns, we need to check if any descs is in tailq */
		if (empty) {
			/* one bdev is empty, check next until all bdev is checked */
			bdev->disk.internal.ns_status = SPDK_BDEV_NS_STATUS_REMOVING;
			pthread_mutex_unlock(&bdev->disk.internal.mutex);
		} else {
			/* means at least one bdev is used, so we just quit this process
			  and mark the status is false. */
			pthread_mutex_unlock(&bdev->disk.internal.mutex);
			break;
		}
	}
	return empty;
}

void spdk_bdev_set_ns_normal(struct nvme_bdev_ctrlr *nvme_bdev_ctrlr, uint32_t nsid)
{
	struct nvme_bdev_ns *ns = NULL;
	struct nvme_bdev *bdev = NULL, *tmp = NULL;

	ns = nvme_bdev_ctrlr->namespaces[nsid - 1];
	if (ns == NULL) {
		return;
	}

	TAILQ_FOREACH_SAFE(bdev, &ns->bdevs, tailq, tmp) {
		pthread_mutex_lock(&bdev->disk.internal.mutex);
		/* set the ns_status to ready case ns delete fail */
		if (bdev->disk.internal.ns_status == SPDK_BDEV_NS_STATUS_REMOVING) {
			bdev->disk.internal.ns_status = SPDK_BDEV_NS_STATUS_READY;
		}
		pthread_mutex_unlock(&bdev->disk.internal.mutex);
	}
}

static void check_error_type(int rc, bool read, void *qpair)
{
	if (rc == -ENOMEM) {
		SPDK_NOTICELOG("%s failed: rc = %d\n", read ? "read" : "write", rc);
	} else if (rc < 0) {
		SPDK_ERRLOG("%s failed: rc = %d, qpair is %p\n", read ? "read" : "write", rc, qpair);
	}
}

int bdev_nvme_queue_cmd_with_md(struct nvme_bdev *bdev, struct spdk_nvme_qpair *qpair,
				void *driver_ctx,
				void *buffer, void *metadata, int direction, uint64_t lba_count, uint64_t lba)
{
	int rc;
	uint32_t io_flags = 0;
	uint8_t *bdev_io_action = (uint8_t *)driver_ctx;
	/* filter bit 0&1 of io->pi_action to get pi_action */
	uint8_t pi_action = bdev_io_action[SPDK_BDEV_IO_ACTION_PI] & 0x03;
	uint8_t dif_flag = bdev_io_action[SPDK_BDEV_IO_ACTION_PI];
	uint8_t fua = bdev_io_action[SPDK_BDEV_IO_ACTION_FUA];
	uint32_t pi_type;

	spdk_bdev_set_io_location(driver_ctx, (uint8_t)LOCAL_LIBSTORAGE_BDEV_NVME);

	if (pi_action > IO_NO_PROTECTION) {
		pi_type = spdk_nvme_ns_get_pi_type(bdev->nvme_ns->ns);
		if (dif_flag & FLAG_PRCHK) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_GUARD;
		}
		/* type3 not support ref tag */
		if (!(dif_flag & FLAG_NO_REF) && (pi_type != SPDK_NVME_FMT_NVM_PROTECTION_TYPE3)) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_REFTAG;
		}
		if (pi_action == IO_HALF_WAY_PROTECTION) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRACT;
		}
	}

	if (fua) {
		io_flags |= SPDK_NVME_IO_FLAGS_FORCE_UNIT_ACCESS;
	}

	if (direction == SPDK_BDEV_IO_TYPE_READ) {
		rc = spdk_nvme_ns_cmd_read_with_md(bdev->nvme_ns->ns, qpair, buffer, metadata, lba,
						   lba_count, bdev_nvme_queued_done, driver_ctx, io_flags, 0, 0);
	} else {
		rc = spdk_nvme_ns_cmd_write_with_md(bdev->nvme_ns->ns, qpair, buffer, metadata, lba,
						    lba_count, bdev_nvme_queued_done, driver_ctx, io_flags, 0, 0);
	}

	check_error_type(rc, direction == SPDK_BDEV_IO_TYPE_READ, qpair);
	return rc;
}

int bdev_nvme_queue_cmd_v_with_md(struct nvme_bdev *bdev, struct spdk_nvme_qpair *qpair,
				  void *driver_ctx,
				  int direction, struct iovec *iov, int iovcnt, uint64_t lba_count, uint64_t lba)
{
	int rc;
	struct nvme_bdev_io *bio = NULL;
	uint32_t io_flags = 0;
	uint8_t *bdev_io_action = (uint8_t *)driver_ctx;
	/* filter bit 0&1 of io->pi_action to get pi_action */
	uint8_t pi_action = bdev_io_action[SPDK_BDEV_IO_ACTION_PI] & 0x03;
	uint8_t dif_flag = bdev_io_action[SPDK_BDEV_IO_ACTION_PI];
	uint8_t fua = bdev_io_action[SPDK_BDEV_IO_ACTION_FUA];
	uint32_t pi_type;

	spdk_bdev_set_io_location(driver_ctx, (uint8_t)LOCAL_LIBSTORAGE_BDEV_NVME);

	if (pi_action > IO_NO_PROTECTION) {
		pi_type = spdk_nvme_ns_get_pi_type(bdev->nvme_ns->ns);
		if (dif_flag & FLAG_PRCHK) {
		    io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_GUARD;
		}
		/* type3 not support ref tag */
		if (!(dif_flag & FLAG_NO_REF) && (pi_type != SPDK_NVME_FMT_NVM_PROTECTION_TYPE3)) {
		    io_flags |= SPDK_NVME_IO_FLAGS_PRCHK_REFTAG;
		}
		if (pi_action == IO_HALF_WAY_PROTECTION) {
			io_flags |= SPDK_NVME_IO_FLAGS_PRACT;
		}
	}

	if (fua) {
		io_flags |= SPDK_NVME_IO_FLAGS_FORCE_UNIT_ACCESS;
	}

	bio = nvme_bdev_io_update_args((struct nvme_bdev_io *)driver_ctx, iov, iovcnt);

	if (direction == SPDK_BDEV_IO_TYPE_READ) {
		rc = spdk_nvme_ns_cmd_readv(bdev->nvme_ns->ns, qpair, lba,
					    lba_count, bdev_nvme_queued_done, bio, io_flags,
					    bdev_nvme_queued_reset_sgl, bdev_nvme_queued_next_sge);
	} else {
		rc = spdk_nvme_ns_cmd_writev(bdev->nvme_ns->ns, qpair, lba, lba_count,
						    bdev_nvme_queued_done, bio, io_flags,
						    bdev_nvme_queued_reset_sgl, bdev_nvme_queued_next_sge);
	}

	check_error_type(rc, direction == SPDK_BDEV_IO_TYPE_READ, qpair);
	return rc;
}

struct nvme_bdev_ctrlr *bdev_nvme_get_ctrlr_by_bdev_desc(void *bdev_desc)
{
	struct spdk_bdev *bdev = spdk_bdev_desc_get_bdev(bdev_desc);
	struct nvme_bdev *nbdev = (struct nvme_bdev *)bdev->ctxt;
	if (nbdev == NULL) {
		return NULL;
	}
	return nbdev->nvme_ns->ctrlr;
}

int bdev_nvme_unmap_blocks(struct nvme_bdev *nbdev, struct spdk_io_channel *ch, void *driver_ctx,
			   struct spdk_nvme_dsm_range *unmap_d, uint32_t unmap_count)
{
	struct nvme_io_channel *nvme_ch = spdk_io_channel_get_ctx(ch);
	uint32_t i;

	if (unmap_count == 0 || unmap_count > SPDK_NVME_DATASET_MANAGEMENT_MAX_RANGES) {
		SPDK_ERRLOG("Invalid parameter, unmap count: %u\n", unmap_count);
		return -EINVAL;
	}

	if (unmap_d == NULL) {
		return -EINVAL;
	}

	for (i = 0; i < unmap_count; i++) {
		if (unmap_d[i].length > SPDK_NVME_DATASET_MANAGEMENT_RANGE_MAX_BLOCKS) {
			SPDK_ERRLOG("Invalid parameter, unmap block count: %u\n", unmap_d[i].length);
			return -EINVAL;
		}
		unmap_d[i].attributes.raw = 0;
	}

	spdk_bdev_set_io_location(driver_ctx, (uint8_t)LOCAL_LIBSTORAGE_BDEV_NVME);
	return spdk_nvme_ns_cmd_dataset_management(nbdev->nvme_ns->ns, nvme_ch->qpair,
			SPDK_NVME_DSM_ATTR_DEALLOCATE,
			unmap_d, unmap_count,
			bdev_nvme_queued_done, driver_ctx);
}

void spdk_bdev_nvme_remove_cb(void *cb_ctx, void *ctrlr)
{
	remove_cb(cb_ctx, (struct spdk_nvme_ctrlr *)ctrlr);
}

void spdk_bdev_fail_ctrlr(void *cb_ctx, void *ctrlr)
{
	spdk_nvme_ctrlr_fail_hotplug((struct spdk_nvme_ctrlr *)ctrlr);
	remove_cb(cb_ctx, (struct spdk_nvme_ctrlr *)ctrlr);
}

struct spdk_nvme_ctrlr *spdk_nvme_bdev_ctrlr_get(char *pci_trid)
{
	struct nvme_bdev_ctrlr	*nvme_bdev_ctrlr;

	TAILQ_FOREACH(nvme_bdev_ctrlr, &g_nvme_bdev_ctrlrs, tailq) {
		if (strcmp(nvme_bdev_ctrlr->connected_trid->traddr, pci_trid) == 0) {
			return nvme_bdev_ctrlr->ctrlr;
		}
	}

	return NULL;
}

int spdk_bdev_nvme_create_self(struct spdk_nvme_transport_id *trid,
			       const char *base_name,
			       const char **names, size_t *count,
			       const char *hostnqn)
{
	struct nvme_probe_ctx	*probe_ctx;
	struct nvme_bdev_ctrlr	*nvme_bdev_ctrlr;
	struct nvme_bdev_ns	*ns;
	struct nvme_bdev	*nvme_bdev;
	struct nvme_bdev	*tmp = NULL;
	uint32_t		i, nsid;
	size_t			j;

	if (nvme_bdev_ctrlr_get(trid) != NULL) {
		SPDK_ERRLOG("A controller with the trid (traddr: %s) already exists.\n", trid->traddr);
		return -1;
	}

	probe_ctx = bdev_nvme_create_probe_ctx(trid, base_name, hostnqn);
	if (probe_ctx == NULL) {
		SPDK_ERRLOG("Failed to create probe_ctx\n");
		return -1;
	}

	if (spdk_nvme_probe(trid, probe_ctx, probe_cb, attach_cb, remove_cb)) {
		SPDK_ERRLOG("Failed to probe for new devices\n");
		free(probe_ctx);
		return -1;
	}

	nvme_bdev_ctrlr = nvme_bdev_ctrlr_get(trid);
	if (!nvme_bdev_ctrlr) {
		SPDK_ERRLOG("Failed to find new NVMe controller\n");
		free(probe_ctx);
		return -1;
	}

	/*
	 * Report the new bdevs that were created in this call.
	 * There can be more than one bdev per NVMe controller since one bdev is created per namespace.
	 */
	j = 0;
	for (i = 0; i < nvme_bdev_ctrlr->num_ns; i++) {
		nsid = i + 1;
		ns = nvme_bdev_ctrlr->namespaces[nsid - 1];

		if (!ns->populated) {
			continue;
		}
		assert(ns->id == nsid);
		TAILQ_FOREACH_SAFE(nvme_bdev, &ns->bdevs, tailq, tmp) {
			if (j < *count) {
				j++;
				names[j] = nvme_bdev->disk.name;
			} else {
				SPDK_ERRLOG("Maximum number of namespaces is %zu.", *count);
				free(probe_ctx);
				return -1;
			}
		}
	}

	*count = j;

	free(probe_ctx);
	return 0;
}
