#include <ocf/ocf.h>

#include "spdk/bdev_module.h"
#include "spdk/thread.h"
#include "spdk/log.h"

#include "vbdev_ocf.h"
#include "data.h"
#include "ctx.h"
#include "volume.h"

#include "vbdev_xcache.h"

static inline struct spdk_bdev_io *spdk_io_to_bdev_io(struct xcache_io *io)
{
	struct spdk_bdev_io *bdev_io = container_of((void *)io - sizeof(struct bdev_ocf_data), struct spdk_bdev_io, driver_ctx);
	return bdev_io;
}

static void
xcache_submit_io_cb(struct xcache_io *io, int error)
{

	struct spdk_bdev_io *bdev_io = spdk_io_to_bdev_io(io);

	if (error == 0) {
		spdk_bdev_io_complete(bdev_io, SPDK_BDEV_IO_STATUS_SUCCESS);
	} else if (error == -ENOMEM) {
		spdk_bdev_io_complete(bdev_io, SPDK_BDEV_IO_STATUS_NOMEM);
	} else {
		spdk_bdev_io_complete(bdev_io, SPDK_BDEV_IO_STATUS_FAILED);
	}
}

static int
submit_io_to_xcache(struct spdk_bdev_io *bdev_io, struct xcache_io *io)
{
	switch (bdev_io->type) {
	case SPDK_BDEV_IO_TYPE_WRITE:
	case SPDK_BDEV_IO_TYPE_READ:
	case SPDK_BDEV_IO_TYPE_FLUSH:
		xcache_submit_io(io);
		return 0;
	default:
		SPDK_ERRLOG("Unsupported IO type: %d\n", bdev_io->type);
		return -EINVAL;
	}
}

static void
xcache_io_handle(struct spdk_io_channel *ch, struct spdk_bdev_io *bdev_io, int dir)
{
	struct vbdev_ocf_qctx *qctx = spdk_io_channel_get_ctx(ch);
	struct vbdev_ocf *vbdev = bdev_io->bdev->ctxt;

	struct bdev_ocf_data *data;
	struct xcache_io *io;
	int err;

	data = vbdev_ocf_data_from_spdk_io(bdev_io);
	io = (struct xcache_io *)(data + 1);
	io->data = data;
	io->io_queue = qctx->queue;
	io->end = xcache_submit_io_cb;
	io->error = 0;
	io->rw = dir;
	io->size = bdev_io->u.bdev.num_blocks * bdev_io->bdev->blocklen;
	io->start_addr = bdev_io->u.bdev.offset_blocks * bdev_io->bdev->blocklen;
	io->core = vbdev->ocf_core;
	io->cache = vbdev->ocf_cache;
	io->flags = 0;

	err = submit_io_to_xcache(bdev_io, io);
	if (err) {
		goto fail;
	}

	return;

fail:
	if (err == -ENOMEM) {
		spdk_bdev_io_complete(bdev_io, SPDK_BDEV_IO_STATUS_NOMEM);
	} else {
		spdk_bdev_io_complete(bdev_io, SPDK_BDEV_IO_STATUS_FAILED);
	}
}

static void
xcache_get_buf_cb(struct spdk_io_channel *ch, struct spdk_bdev_io *bdev_io,
		     bool success)
{
	if (!success) {
		spdk_bdev_io_complete(bdev_io, SPDK_BDEV_IO_STATUS_FAILED);
		return;
	}

	xcache_io_handle(ch, bdev_io, OCF_READ);
}

/* Called from bdev layer when an io to Cache vbdev is submitted */
void
xcache_submit_request(struct spdk_io_channel *ch, struct spdk_bdev_io *bdev_io)
{
	switch (bdev_io->type) {
	case SPDK_BDEV_IO_TYPE_READ:
		/* User does not have to allocate io vectors for the request,
		 * so in case they are not allocated, we allocate them here */
		spdk_bdev_io_get_buf(bdev_io, xcache_get_buf_cb,
				     bdev_io->u.bdev.num_blocks * bdev_io->bdev->blocklen);
		break;
	case SPDK_BDEV_IO_TYPE_WRITE:
		xcache_io_handle(ch, bdev_io, OCF_WRITE);
		break;
	case SPDK_BDEV_IO_TYPE_FLUSH:
		xcache_io_handle(ch, bdev_io, OCF_FLUSH);
		break;
	default:
		SPDK_ERRLOG("Unknown I/O type %d\n", bdev_io->type);
		spdk_bdev_io_complete(bdev_io, SPDK_BDEV_IO_STATUS_FAILED);
		break;
	}
}

int
xcache_get_ctx_size(void)
{
	return sizeof(struct xcache_io) + sizeof(struct bdev_ocf_data);
}

static void
backdev_io_end(bool success, struct xcache_backdev_io *io_base,
		uint64_t addr, uint64_t size)
{
	struct backdev_io_end_arg cb_arg;

	if (io_base->priv != NULL) {
		env_free(io_base->priv);
		io_base->priv = NULL;
	}

	cb_arg.error = !success;
	cb_arg.addr = addr;
	cb_arg.size = size;
	xcache_backdev_io_end(io_base, &cb_arg);
}

static void
backdev_io_cb(struct spdk_bdev_io *bdev_io, bool success, void *opaque)
{
	uint64_t addr = bdev_io->u.bdev.offset_blocks * bdev_io->bdev->blocklen;
	uint64_t size = bdev_io->u.bdev.num_blocks * bdev_io->bdev->blocklen;

	backdev_io_end(success, (struct xcache_backdev_io *)opaque, addr, size);

	spdk_bdev_free_io(bdev_io);
}

static int xcache_prepare_submit(struct xcache_backdev_io *io_base, bool to_cache,
		struct vbdev_ocf_base **base, struct spdk_io_channel **ch)
{
	struct xcache_io *io = io_base->xcache_io;
	ocf_queue_t q = io->io_queue;
	ocf_cache_t cache = ocf_queue_get_cache(q);
	struct vbdev_ocf_cache_ctx *cctx = ocf_cache_get_priv(cache);
	struct vbdev_ocf *vbdev;
	struct vbdev_ocf_qctx *qctx;

	if (cctx == NULL) {
		return -EFAULT;
	}

	vbdev = cctx->vbdev;
	// get vbdev_ocf_base
	if (to_cache) {
		*base = &vbdev->cache;
	} else {
		*base = &vbdev->core;
	}

	if (q == cctx->cleaner_queue || q == cctx->mngt_queue) {
		*ch = (*base)->management_channel;
		return 0;
	}

	qctx = ocf_queue_get_priv(q);
	if (qctx == NULL) {
		return -EFAULT;
	}
	if (to_cache) {
		*ch = qctx->cache_ch;
	} else {
		*ch = qctx->core_ch;
	}

	return 0;
}

struct copy_vector {
	int iovcnt;
	struct iovec iovs[];
};

void spdk_backdev_submit_io(struct xcache_backdev_io *io_base, bool to_cache, uint64_t addr, uint64_t len, uint64_t offset, uint8_t dir)
{
	struct bdev_ocf_data *data = (struct bdev_ocf_data *)io_base->data;

	uint64_t size, bytes;
	struct iovec *iovs;
	int iovcnt;
	int status = 0;
	struct vbdev_ocf_base *base;
	struct spdk_io_channel *ch;
	int i, j;
	struct copy_vector *vector;

	io_base->priv = NULL;

	if (xcache_prepare_submit(io_base, to_cache, &base, &ch) != 0) {
		backdev_io_end(false, io_base, addr, len);
		return;
	}

	if (dir == OCF_FLUSH) {
		status = spdk_bdev_flush(base->desc, ch, addr, len, backdev_io_cb, io_base);
		goto end;
	}

	size = data->size;
	iovs = data->iovs;
	iovcnt = data->iovcnt;
	if (len < size) {
		if (iovcnt == 1) {
			if (dir == OCF_READ) {
				status = spdk_bdev_read(base->desc, ch,
							iovs[0].iov_base + offset, addr, len,
							backdev_io_cb, io_base);
			} else if (dir == OCF_WRITE) {
				status = spdk_bdev_write(base->desc, ch,
							 iovs[0].iov_base + offset, addr, len,
							 backdev_io_cb, io_base);
			}
			goto end;
		} else {
			i = get_starting_vec(iovs, iovcnt, &offset);

			if (i < 0) {
				SPDK_ERRLOG("offset bigger than data size\n");
				backdev_io_end(false, io_base, addr, len);
				return;
			}

			bytes = len + offset;
			j = get_starting_vec(iovs + i, iovcnt - i, &bytes);
			if (offset == 0 && bytes == 0) {
				iovs = iovs + i;
				iovcnt = j;
			} else {
				if (bytes != 0) {
					j++;
				}
				vector = env_malloc(sizeof(int) + sizeof(*iovs) * j, ENV_MEM_NOIO);
				if (vector == NULL) {
					SPDK_ERRLOG("allocation failed\n");
					backdev_io_end(false, io_base, addr, len);
					return;
				}
				vector->iovcnt = j;
				io_base->priv = vector;

				initialize_cpy_vector(vector->iovs, &iovs[i], offset, len);

				iovs = vector->iovs;
				iovcnt = vector->iovcnt;
			}
		}
	}

	if (dir == OCF_READ) {
		status = spdk_bdev_readv(base->desc, ch,
					 iovs, iovcnt, addr, len, backdev_io_cb, io_base);
	} else if (dir == OCF_WRITE) {
		status = spdk_bdev_writev(base->desc, ch,
					  iovs, iovcnt, addr, len, backdev_io_cb, io_base);
	}

end:
	if (status) {
		SPDK_ERRLOG("submission failed with status=%d\n", status);
		backdev_io_end(false, io_base, addr, len);
	}
}
